from vai_lab._plugin_templates import UI
from vai_lab._import_helper import get_lib_parent_dir
from vai_lab._types import DictT, DataInterface, GUICoreInterface

import os
import numpy as np
import pandas as pd
from typing import Tuple, List, Union
from PIL import Image, ImageTk, PngImagePlugin

import tkinter as tk
from tkinter import messagebox, ttk
from tkinter.filedialog import asksaveasfile

_PLUGIN_READABLE_NAMES = {"manual": "default",
                          "binary": "alias",
                          "classification": "alias"}            # type:ignore
_PLUGIN_MODULE_OPTIONS = {"layer_priority": 2,
                          "required_children": None}            # type:ignore
_PLUGIN_REQUIRED_SETTINGS = {"image_dir": "str"}                # type:ignore
_PLUGIN_OPTIONAL_SETTINGS = {}                                  # type:ignore
_PLUGIN_REQUIRED_DATA = {"X", "Y"}                              # type:ignore


class ManualInput(tk.Frame, UI):            # type:ignore
    """Method of user interaction for binary or classification data"""

    def __init__(self, parent, controller, config: DictT):
        self.parent = parent
        super().__init__(parent, bg=self.parent['bg'])
        self.controller: GUICoreInterface = controller
        self.controller.title('Manual Input')

        self.dirpath = get_lib_parent_dir()

        
        self.assets_path = os.path.join(self.dirpath, 'utils', 'resources', 'Assets')

        self._data_in: DataInterface
        self._class_list = None
        self._config = config
        self.save_path = ''
        self.saved = True
        

    def _load_images_from_data(self):
        self.image_list: List[Union[ImageTk.PhotoImage,Image.Image]] = []
        pixels = 500
        for f in self._data_in["X"].keys():
            self.image_list.append(
                ImageTk.PhotoImage(
                    self.expand2square(Image.fromarray(self._data_in["X"][f]), (0, 0, 0))
                    .resize((pixels, pixels))))

        # Inital window
        self.my_label = tk.Label(self, image=self.image_list[0],
                                 bg=self.parent['bg'])
        self.my_label.grid(column=0, row=0, rowspan=10, columnspan=3)
        self.N = len(self._data_in["X"])

        # Status bar in the lower part of the window
        status = tk.Label(self, text='Image 1 of '+str(self.N), bd=1,
                          relief=tk.SUNKEN, anchor=tk.E, fg='white',
                          bg=self.parent['bg'])
        status.grid(row=20, column=0, columnspan=4, pady=10,
                    sticky=tk.W+tk.E)

        self.frame1 = tk.Frame(self, bg=self.parent['bg'])
        frame4 = tk.Frame(self, bg=self.parent['bg'])
        frame5 = tk.Frame(self, bg=self.parent['bg'])
        frame6 = tk.Frame(self, bg=self.parent['bg'])

        # Buttons initialisation
        self.back_img = ImageTk.PhotoImage(Image.open(
            os.path.join(self.assets_path, 'back_arrow.png')).resize((150, 50)))
        self.forw_img = ImageTk.PhotoImage(Image.open(
            os.path.join(self.assets_path, 'forw_arrow.png')).resize((150, 50)))
        self.button_back = tk.Button(
            frame4, image=self.back_img, bg=self.parent['bg'],
            state=tk.DISABLED)
        self.button_back.grid(column=0, row=19, padx=(
            10, 0), pady=10, sticky="news")
        self.button_save = tk.Button(
            frame4, text='Save', fg='white', bg=self.parent['bg'], height=3,
            width=20, command=self.save_file)
        self.button_save.grid(column=1, row=19, sticky="news", pady=10)
        self.button_forw = tk.Button(
            frame4, image=self.forw_img, bg=self.parent['bg'],
            command=lambda: self.forward_back(2))
        self.button_forw.grid(column=2, row=19, sticky="news", pady=10)

        tk.Button(
            frame5, text="Done",
            fg='white', bg=self.parent['bg'],
            height=3, width=20,
            command=self.check_quit).grid(column=4, row=19, sticky="news", pady=10)

        img = self.image_list[0]
        self.my_img = tk.Label(self.frame1, image=img,
                               bg=self.parent['bg'])
        self.my_img.pack(fill=tk.BOTH, expand=True, padx=(10, 0), pady=(10, 0))
        self.my_img.bind("<Configure>", self.resizing)

        self.frame1.grid(row=0, column=0, sticky="nsew")
        # self.frame3.grid(row = 0, column = 2, sticky="nsew", pady=10, padx=10)
        frame4.grid(row=1, column=0, sticky="sew")
        frame5.grid(row=1, column=1, sticky="sew")
        frame6.grid(row=2, column=0, columnspan=3, sticky="sew")

        frame4.grid_columnconfigure(tuple(range(3)), weight=1)
        self.grid_rowconfigure(0, weight=3)
        self.grid_rowconfigure(1, weight=1)
        self.grid_columnconfigure(tuple(range(3)), weight=1)
        self.grid_columnconfigure(0, weight=2)

    def set_data_in(self, data_in):
        req_check = [
            r for r in _PLUGIN_REQUIRED_DATA if r not in data_in.keys()]
        if len(req_check) > 0:
            raise Exception("Minimal Data Requirements not met"
                            + "\n\t{0} ".format("ManualInput")
                            + "requires data: {0}".format(_PLUGIN_REQUIRED_DATA)
                            + "\n\tThe following data is missing:"
                            + "\n\t\u2022 {}".format(",\n\t\u2022 ".join([*req_check])))
        self._data_in = data_in
        self._load_images_from_data()
        self._load_classes_from_data()

    def class_list(self):
        """Getter for required _class_list variable

        :return: list of class labels
        :rtype: list of strings
        """
        return self._class_list

    def _load_classes_from_data(self):
        """Setter for required _class_list variable

        :param value: class labels for binary classification
        :type value: list of strings
        """
        self._class_list = list(self._data_in["Y"]["Class"])
        self.out_data = np.zeros((self.N, len(self._class_list)))

        frame2 = tk.Frame(self, bg=self.parent['bg'])
        frame2.grid(row=0, column=1, sticky="nsew")

        self.button_cl = {}

        self.var = {}
        for i, cl in enumerate(self._class_list):
            self.var[0, i] = tk.IntVar(value=self.out_data[0, i])
            self.button_cl[cl] = tk.Checkbutton(
                frame2, text=cl, fg='white', bg=self.parent['bg'],
                selectcolor='black', height=3, width=20,
                variable=self.var[0, i],
                command=(lambda i=i: self.onPress(0, i)))
            self.button_cl[cl].grid(column=4, row=i)

        # Tree defintion. Output display
        style = ttk.Style()
        style.configure(
            "Treeview", background='white', foreground='white',
            rowheight=25, fieldbackground='white',
            font=self.controller.pages_font)
        style.configure("Treeview.Heading", font=self.controller.pages_font)
        style.map('Treeview', background=[('selected', 'grey')])

        tree_frame = tk.Frame(self)
        tree_frame.grid(row=0, column=2, sticky="nsew", pady=10, padx=10)

        tree_scrollx = tk.Scrollbar(tree_frame, orient='horizontal')
        tree_scrollx.pack(side=tk.BOTTOM, fill=tk.X)
        tree_scrolly = tk.Scrollbar(tree_frame)
        tree_scrolly.pack(side=tk.RIGHT, fill=tk.Y)

        self.tree = ttk.Treeview(tree_frame,
                                 yscrollcommand=tree_scrolly.set,
                                 xscrollcommand=tree_scrollx.set)
        self.tree.pack(fill='both', expand=True)

        tree_scrollx.config(command=self.tree.xview)
        tree_scrolly.config(command=self.tree.yview)

        self.tree['columns'] = self._class_list

        # Format columns
        self.tree.column("#0", width=80,
                         minwidth=50)
        for n, cl in enumerate(self._class_list):
            self.tree.column(
                cl, width=int(self.controller.pages_font.measure(str(cl)))+20,
                minwidth=50, anchor=tk.CENTER)
        # Headings
        self.tree.heading("#0", text="Image", anchor=tk.CENTER)
        for cl in self._class_list:
            self.tree.heading(cl, text=cl, anchor=tk.CENTER)
        self.tree.tag_configure('odd', foreground='black',
                                background='#E8E8E8')
        self.tree.tag_configure('even', foreground='black',
                                background='#DFDFDF')
        # Add data
        for n, sample in enumerate(self.out_data):
            if n % 2 == 0:
                self.tree.insert(parent='', index='end', iid=n, text=n+1,
                                 values=tuple(sample.astype(int)), tags=('even',))
            else:
                self.tree.insert(parent='', index='end', iid=n, text=n+1,
                                 values=tuple(sample.astype(int)), tags=('odd',))

        # Select the current row
        self.tree.selection_set(str(int(0)))

        # Define double-click on row action
        self.tree.bind("<Double-1>", self.OnDoubleClick)

    def expand2square(self,
                        pil_img :PngImagePlugin.PngImageFile,
                        background_color :Tuple):
        " Adds padding to make the image a square."
        width, height = pil_img.size
        if width == height:
            return pil_img
        elif width > height:
            result = Image.new(pil_img.mode, (width, width), background_color)
            result.paste(pil_img, (0, (width - height) // 2))
            return result
        else:
            result = Image.new(pil_img.mode, (height, height),
                               background_color)
            result.paste(pil_img, ((height - width) // 2, 0))
            return result

    def resizing(self, event):
        """ Resizes window to tree view height and buttons width """
        n = int(self.tree.selection()[0])
        iw, ih = self.image_list[n].width, self.image_list[n].height
        iw = iw() if type(iw) is not int else iw
        ih = ih() if type(ih) is not int else ih

        mw, mh = self.frame1.winfo_width()-14, self.frame1.winfo_height() - \
            14  # Frame border correction
        if (iw != mw) and (ih != mh):
            if iw > ih:
                ih = ih*(mw/iw)
                r = mh/ih if (ih/mh) > 1 else 1
                iw, ih = mw*r, ih*r
            else:
                iw = iw*(mh/ih)
                r = mw/iw if (iw/mw) > 1 else 1
                iw, ih = iw*r, mh*r
            
            if type(self.image_list[n]) == ImageTk.PhotoImage:
                _temp_img:Image = ImageTk.getimage(self.image_list[n])
            else:
                _temp_img:Image = self.image_list[n]

            self.image_list[n] = ImageTk.PhotoImage(_temp_img.resize((int(iw), int(ih))))
        self.my_img.config(image=self.image_list[n])

    def check_quit(self):

        if not self.saved:
            response = messagebox.askokcancel(
                "Exit?",
                "Do you want to leave the program without saving?")
            if response:
                self.controller.destroy()
        else:
            response = messagebox.askokcancel(
                "Exit?",
                "Are you sure you are finished?")
            self.controller.destroy()

    def save_file_as(self):

        self.save_path = asksaveasfile(mode='w')
        self.save_file()

    def save_file(self):

        if self.save_path == '':
            self.save_path = asksaveasfile(defaultextension='.txt',
                                           filetypes=[('Text file', '.txt'),
                                                      ('CSV file', '.csv'),
                                                      ('All Files', '*.*')])
        # asksaveasfile return `None` if dialog closed with "cancel".
        if self.save_path is not None:
            filedata = pd.DataFrame(
                self.out_data, columns=self._class_list).to_string()
            self.save_path.seek(0)  # Move to the first row to overwrite it
            self.save_path.write(filedata)
            self.save_path.flush()  # Save without closing
            # typically the above line would do. however this is used to ensure that the file is written
            os.fsync(self.save_path.fileno())
            self.saved = True

    def forward_back(self, image_number):
        " Forward button to continue to the next image in the folder."

        self.tree.selection_set(str(int(image_number-1)))
        # Print the corresponding image
        self.resizing((0, 0))

        # Update button commands
        self.button_forw.config(image=self.forw_img, bg=self.parent['bg'],
                                command=lambda: self.forward_back(
                                    image_number+1),
                                text='', state=tk.NORMAL)
        self.button_back.config(image=self.back_img, bg=self.parent['bg'],
                                command=lambda: self.forward_back(
                                    image_number-1),
                                text='', state=tk.NORMAL)
        if image_number == self.N:
            self.button_forw.config(state=tk.DISABLED)
        if image_number == 1:
            self.button_back.config(state=tk.DISABLED)

        # Classes buttons
        # var = {}
        for i, cl in enumerate(self._class_list):
            # print(out_data[image_number-1,i])
            self.var[image_number-1, i] = tk.IntVar(
                value=int(self.out_data[image_number-1, i]))
            # var[i] = tk.IntVar(value = int(self.out_data[image_number-1,i]))
            # I can not make this be selected when going backwards or forward
            # if it was previously selected.
            self.button_cl[cl].config(text=cl, fg='white', bg=self.parent['bg'],
                                      selectcolor='black', height=3, width=20,
                                      variable=self.var[image_number-1, i],
                                      command=(lambda i=i: self.onPress(image_number-1, i)))

        # # Status bar
        # status = tk.Label(
        #     self, text='Image ' + str(image_number) + ' of '+str(self.N),
        #     bd = 1, relief = tk.SUNKEN, anchor = tk.E, fg = 'white',
        #     bg = self.parent['bg'])
        # status.grid(row=20, column=0, columnspan=4, pady = 10,
        #             sticky = tk.W+tk.E)

    def onPress(self, n, i):
        "Updates the stored values on clicking the checkbutton."

        self.out_data[n, i] = not self.out_data[n, i]
        self.tree.item(self.tree.get_children()[n], text=n+1,
                       values=tuple(self.out_data[n, :].astype(int)))
        self.saved = False

    def OnDoubleClick(self, event):
        "Moves to the image corresponding to the row clicked on the tree."

        item = self.tree.selection()[0]
        self.forward_back(self.tree.item(item, "text"))



# -*- coding: utf-8 -*-

"""Import all submodules inside this folder and expose them with their respective names

Let's look for a more automatic way of doing this when we have more modules
"""
# from .User_Feedback_template import UI
# from .GUI_core import GUI

# from .. import *


from typing import Any, Protocol, KeysView, Dict, TypeVar
from pandas.core.frame import DataFrame
from tkinter.font import Font

DataInterfaceT = TypeVar("DataInterfaceT", bound="DataInterface")
DictT = Dict[str, Dict]


class PluginOptions(Protocol):
    _PLUGIN_READABLE_NAMES: dict
    _PLUGIN_MODULE_OPTIONS: dict
    _PLUGIN_REQUIRED_SETTINGS: dict
    _PLUGIN_OPTIONAL_SETTINGS: dict
    _PLUGIN_REQUIRED_DATA: dict
    _PLUGIN_OPTIONAL_DATA: dict


class DataInterface(Protocol):
    def __init__(self) -> None:
        ...

    def import_data(self, config: dict) -> None:
        ...

    def import_data_from_config(self, config: dict) -> None:
        ...

    def append_data_column(self, col_name: str, data=None) -> None:
        ...

    def keys(self) -> KeysView:
        ...

    def copy(self: DataInterfaceT) -> DataInterfaceT:
        ...

    def __getitem__(self, key: str) -> DataFrame:
        ...


class PluginInterface(Protocol):
    def configure(self, config: dict) -> None:
        ...

    def set_data_in(self, data_in: DataInterface) -> None:
        ...


class DataProcessingPluginInterface(PluginInterface, Protocol):
    def fit(self):
        ...

    def transform(self, data: DataInterfaceT) -> DataInterface:
        ...

class EnvironmentPluginInterface(PluginInterface, Protocol):
    def load_model(self) -> None:
        ...

    def connect(self):
        ...

    def reset(self):
        ...

    def disconnect(self):
        ...

    def run_simulation(self):
        ...


class PluginSpecsInterface(Protocol):
    @property
    def names(self):
        ...

    @property
    def class_names(self):
        ...

    @property
    def module_options(self):
        ...

    @property
    def required_settings(self):
        ...

    @property
    def class_descriptions(self):
        ...

    @property
    def optional_settings(self):
        ...

    @property
    def available_plugin_names(self):
        ...

    def find_from_class_name(self, value):
        ...

    def find_from_readable_name(self, value):
        ...

    def print(self, value):
        ...


class ModuleInterface(Protocol):
    def set_avail_plugins(self, avail_plugins: PluginSpecsInterface):
        ...

    def set_data_in(self, data_in: DataInterface):
        ...

    def set_options(self, module_config: DictT):
        ...

    def launch(self):
        ...

    def get_result(self) -> DataInterface:
        ...

class GUICoreInterface(ModuleInterface,Protocol):
    title: Any #mypy bug prevents proper typing
    pages_font: Font

    def destroy(self) -> None:
        ...

    def set_gui(self) -> None:
        ...

class InputDataCoreInterface(ModuleInterface,Protocol):
    def load_data_from_file(self, filename:str, data_id:str) -> None:
        ...


from vai_lab.Data.xml_handler import XML_handler
from vai_lab._plugin_helpers import PluginSpecs
from vai_lab._import_helper import get_lib_parent_dir

import os
import numpy as np
import pandas as pd

from typing import Dict, List
from PIL import Image, ImageTk
import tkinter as tk
from tkinter import messagebox, ttk


_PLUGIN_READABLE_NAMES = {"plugin_canvas": "default",
                          "pluginCanvas": "alias",
                          "plugin canvas": "alias"}         # type:ignore
_PLUGIN_MODULE_OPTIONS = {"layer_priority": 2,
                          "required_children": None}        # type:ignore
_PLUGIN_REQUIRED_SETTINGS = {}                              # type:ignore
_PLUGIN_OPTIONAL_SETTINGS = {}                              # type:ignore


class pluginCanvas(tk.Frame):
    """Canvas for graphical specification of plugins"""

    def __init__(self, parent, controller, config: dict):
        """ Here we define the frame displayed for the plugin specification."""

        super().__init__(parent, bg=parent['bg'])
        self.bg = parent['bg']
        self.parent = parent
        self.controller = controller
        self.s = XML_handler()

        script_dir = get_lib_parent_dir()
        self.tk.call('wm', 'iconphoto', self.controller._w, ImageTk.PhotoImage(
            file=os.path.join(os.path.join(
                script_dir,
                'utils',
                'resources',
                'Assets',
                'VAILabsIcon.ico'))))
        # self.grid_rowconfigure(tuple(range(5)), weight=1)
        self.grid_rowconfigure(4, weight=1)
        self.grid_columnconfigure(0, weight=1)

        frame1 = tk.Frame(self, bg=self.bg)
        self.frame2 = tk.Frame(self, bg=self.bg)
        self.framex = tk.Frame(self, bg=self.bg)
        frame3 = tk.Frame(self, bg=self.bg)
        self.frame4 = tk.Frame(self, bg=self.bg)

        # Create canvas
        self.width, self.height = 700, 700
        self.canvas = tk.Canvas(frame1, width=self.width,
                                height=self.height, background="white")
        self.canvas.pack(fill=tk.BOTH, expand=True, padx=(10, 0), pady=10)

        self.m: int
        self.w, self.h = 100, 50
        self.cr = 4
        self.canvas.bind('<Button-1>', self.on_click)
        self.id_done = [0, 1]
        self.id_mod:List[int] = [0, 1]
        self.out_data = pd.DataFrame()
        self.plugin: Dict[int, tk.StringVar] = {}
        self.dataType: Dict[int, tk.StringVar] = {}
        self.allWeHearIs: List[tk.Radiobutton] = []

        self.my_label = tk.Label(self.frame2,
                                 text='',
                                 pady=10,
                                 font=self.controller.title_font,
                                 bg=self.bg,
                                 fg='white',
                                 anchor=tk.CENTER)
        self.my_label.grid(column=5,
                           row=0, columnspan=2, padx=10)

        self.back_img = ImageTk.PhotoImage(Image.open(
            os.path.join(script_dir,
                         'utils',
                         'resources',
                         'Assets',
                         'back_arrow.png')).resize((140, 60)))
        self.forw_img = ImageTk.PhotoImage(Image.open(
            os.path.join(script_dir,
                         'utils',
                         'resources',
                         'Assets',
                         'forw_arrow.png')).resize((140, 60)))

        ww = int(self.width/40)
        tk.Button(
            frame3, text='Load Pipeline', fg='white', bg=parent['bg'],
            height=3, width=ww, font=self.controller.pages_font,
            command=self.upload).grid(column=0, row=0, sticky="news",
                                      padx=(10, 0), pady=(0, 10))
        tk.Button(
            frame3, text='Back to main', fg='white', bg=parent['bg'],
            height=3, width=ww, font=self.controller.pages_font,
            command=self.check_quit).grid(column=1, row=0, sticky="news", pady=(0, 10))

        self.frame_canvas = tk.Canvas(
            self.framex, bg=self.bg, bd=0, highlightthickness=0)
        self.tree_scrolly = tk.Scrollbar(
            self.framex, command=self.frame_canvas.yview)
        self.tree_scrolly.pack(side=tk.RIGHT, fill=tk.Y)
        self.tree_scrollx = tk.Scrollbar(
            self.framex, command=self.frame_canvas.xview, orient='horizontal')
        self.tree_scrollx.pack(side=tk.BOTTOM, fill=tk.X)

        self.save_path = ''
        self.saved = True
        frame1.grid(column=0, row=0, rowspan=5, sticky="nsew")
        self.frame2.grid(column=1, row=0, sticky="new", padx=(10, 0))
        frame3.grid(column=0, row=5, sticky="nswe")
        self.frame4.grid(column=1, row=5, sticky="nsew")

        # self.frame2.grid_columnconfigure(tuple(range(2)), weight=1)
        self.frame2.grid_columnconfigure(1, weight=1)
        frame3.grid_columnconfigure(tuple(range(2)), weight=2)
        # frame3.grid_columnconfigure(2, weight=2)
        self.frame4.grid_columnconfigure(tuple(range(2)), weight=2)
        # self.frame4.grid_columnconfigure(2, weight=2)

    def class_list(self, value):
        """ Temporary fix """
        return value

    def on_click(self, event):
        """ Passes the mouse click coordinates to the select function."""
        self.select(event.x, event.y)

    def select(self, x: float, y: float):
        """ 
        Selects the module at the mouse location and updates the associated 
        plugins as well as the colours. 
        Blue means no plugin has been specified,
        Orange means the module is selected.
        Green means the plugin for this module is already set. 
        :param x: float type of module x coordinate
        :param y: float type of module y coordinate
        """
        self.selected = self.canvas.find_overlapping(
            x-5, y-5, x+5, y+5)
        self.canvas_selected: int
        if self.selected:
            if len(self.selected) > 2:
                self.canvas_selected = self.selected[-2]
            else:
                self.canvas_selected = self.selected[-1]

        self.check_updated()

        if self.m in self.id_done and self.m > 1:
            self.canvas.itemconfig('p'+str(self.m), fill='#46da63')
        else:
            self.canvas.itemconfig('p'+str(self.m), fill=self.bg)

        if len(self.canvas.gettags(self.canvas_selected)) > 0:
            if not (len(self.canvas.gettags(self.canvas_selected)[0].split('-')) > 1) and\
                    not (self.canvas.gettags(self.canvas_selected)[0].split('-')[0] == 'loop'):
                self.m = int(self.canvas.gettags(self.canvas_selected)[0][1:])
            if self.m > 1:
                if self.m not in self.id_done and self.m > 1:
                    self.canvas.itemconfig('p'+str(self.m), fill='#dbaa21')
                for widget in self.allWeHearIs:
                    widget.grid_remove()

                self.display_buttons()
                module_number = self.id_mod.index(self.m)
                if not hasattr(self, 'button_forw'):
                    if module_number == len(self.id_mod)-1:
                        self.button_forw = tk.Button(
                            self.frame4, text='Finish', bg=self.bg,
                            font=self.controller.pages_font,
                            fg='white', height=3, width=15,
                            command=self.finish, state=tk.NORMAL, image='')
                    else:
                        pCoord = self.canvas.coords(
                            'p'+str(self.id_mod[module_number+1]))
                        self.button_forw = tk.Button(
                            self.frame4, image=self.forw_img, bg=self.bg,
                            command=lambda: self.select(
                                pCoord[0], pCoord[1]), state=tk.NORMAL)
                    self.button_forw.grid(
                        column=1, row=0, sticky="news", pady=(0, 10), padx=(0, 10))
                    if module_number < 3:
                        self.button_back = tk.Button(
                            self.frame4, image=self.back_img, bg=self.bg,
                            state=tk.DISABLED)
                    else:
                        mCoord = self.canvas.coords(
                            'p'+str(self.id_mod[module_number-1]))
                        self.button_back = tk.Button(
                            self.frame4, image=self.back_img, bg=self.bg,
                            command=lambda: self.select(
                                mCoord[0], mCoord[1]), state=tk.NORMAL)
                    self.button_back.grid(
                        column=0, row=0, sticky="news", pady=(0, 10))
                else:
                    if module_number == len(self.id_mod)-1:
                        self.button_forw.config(text='Finish', bg=self.bg,
                                                font=self.controller.pages_font,
                                                fg='white', height=3, width=15,
                                                command=self.finish, state=tk.NORMAL, image='')
                    else:
                        pCoord = self.canvas.coords(
                            'p'+str(self.id_mod[module_number+1]))
                        self.button_forw.config(image=self.forw_img, bg=self.bg,
                                                command=lambda: self.select(
                                                    pCoord[0], pCoord[1]), text='', state=tk.NORMAL)
                    if module_number < 3:
                        self.button_back.config(image=self.back_img, bg=self.bg,
                                                state=tk.DISABLED, text='')
                    else:
                        mCoord = self.canvas.coords(
                            'p'+str(self.id_mod[module_number-1]))
                        self.button_back.config(image=self.back_img, bg=self.bg,
                                                command=lambda: self.select(
                                                    mCoord[0], mCoord[1]), text='', state=tk.NORMAL)
            else:  # If user clicks on Initialiser or Output
                self.my_label.config(text='')
                for widget in self.allWeHearIs:
                    widget.grid_remove()
                for widget in self.radio_label:
                    widget.grid_remove()
                if hasattr(self, 'button_forw'):
                    self.button_back.config(image=self.back_img, bg=self.bg,
                                            state=tk.DISABLED, text='')
                    pCoord = self.canvas.coords('p'+str(self.id_mod[2]))
                    self.button_forw.config(image=self.forw_img, bg=self.bg,
                                            command=lambda: self.select(
                                                pCoord[0], pCoord[1]), text='', state=tk.NORMAL)

    def finish(self):
        """ Calls function check_quit.
        Before that, it checks if the current module plugins have been changed 
        and, if so, updates their information in the XML_handler class.
        """
        self.check_updated()
        self.check_quit()

    def check_updated(self):
        """ Checks if the current plugin exists and 
        stores/updates the plugin options
        """
        if (self.m in self.plugin.keys()) and\
                (self.plugin[self.m].get() != 'None'):  # add
            self.id_done.append(self.m)
            self.s.append_plugin_to_module(self.plugin[self.m].get(),
                                           {**self.req_settings, **
                                               self.opt_settings},
                                           np.array(self.module_names)[
                self.m == np.array(self.id_mod)][0],
                True)

    def display_buttons(self):
        """ Updates the displayed radiobuttons and the description windows.
        It loads the information corresponding to the selected module (self.m)
        and shows the available plugins and their corresponding descriptions.
        """
        module = np.array(self.module_list)[self.m == np.array(self.id_mod)][0]
        name = self.canvas.itemcget('t'+str(self.m), 'text')
        ps = PluginSpecs()
        plugin_list = list(ps.class_names[module].values())
        plugin_list.append('Custom')
        descriptions = list(ps.class_descriptions[module].values())
        descriptions.append('User specified plugin')
        type_list = np.array([x.get('Type', None)
                             for x in list(ps.module_options[module].values())])
        type_list = np.append(type_list, None)
        text = ''
        if module.lower() == 'modelling':
            text = '\nPluggins in white correspond to '+self.controller.output_type
        self.my_label.config(
            text='Choose a plugin for the '+name+' module'+text)
        if self.m not in self.plugin:
            self.plugin[self.m] = tk.StringVar()
            self.plugin[self.m].set(None)
        self.allWeHearIs = []

        try:
            for frame in self.framexx:
                frame.grid_forget()
                frame.destroy()
            self.framexx = []
        except AttributeError:
            self.framexx = []
        framexx_name = np.array([])
        framexx_i = []
        self.radio_label = []
        values = np.append(np.unique(type_list[type_list != None]), 'other') if not any(
            type_list == 'other') else np.unique(type_list[type_list != None])
        for i, t in enumerate(values):
            self.framexx.append(tk.Frame(self.framep, bg=self.bg))
            framexx_name = np.append(
                framexx_name, t if t is not None else 'other')
            framexx_i.append(0)
            self.radio_label.append(tk.Label(self.framexx[-1],
                             text=framexx_name[-1].capitalize(),
                             pady=10,
                             font=self.controller.title_font,
                             bg=self.bg,
                             fg='white'))
            self.radio_label[-1].grid(column=0,
                       row=0, padx=(10, 0), sticky='nw')
            self.framexx[-1].grid(column=0, row=i, sticky="nsew")
        unsupervised = ['clustering', 'RL']
        for p, plug in enumerate(plugin_list):
            if module.lower() == 'modelling' and p < len(plugin_list)-1:
                colour = 'white' if type_list[p] == self.controller.output_type or (
                    self.controller.output_type == 'unsupervised' and type_list[p] in unsupervised) else 'grey'
            else:
                colour = 'white'
            frame_idx = np.where(framexx_name == type_list[p])[
                0][0] if type_list[p] in framexx_name else np.where(framexx_name == 'other')[0][0]
            rb = tk.Radiobutton(self.framexx[frame_idx], text=plug, fg=colour, bg=self.bg,
                                height=3, var=self.plugin[self.m],
                                selectcolor='black', value=plug, justify=tk.LEFT,
                                font=self.controller.pages_font, command=self.optionsWindow)
            rb.grid(column=int(framexx_i[frame_idx] % 2 != 0),
                    row=int(framexx_i[frame_idx]/2)+1,
                    sticky='w',
                    padx=(10, 0))
            framexx_i[frame_idx] += 1
            self.CreateToolTip(rb, text=descriptions[p])
            self.allWeHearIs.append(rb)

    def optionsWindow(self):
        """ Function to create a new window displaying the available options 
        of the selected plugin."""

        module = np.array(self.module_list)[self.m == np.array(self.id_mod)][0]
        ps = PluginSpecs()
        self.opt_settings = ps.optional_settings[module][self.plugin[self.m].get(
        )]
        self.req_settings = ps.required_settings[module][self.plugin[self.m].get(
        )]
        if (len(self.opt_settings) != 0) or (len(self.req_settings) != 0):
            if hasattr(self, 'newWindow') and (self.newWindow != None):
                self.newWindow.destroy()
            self.newWindow = tk.Toplevel(self.controller)
            # Window options
            self.newWindow.title(self.plugin[self.m].get()+' plugin options')
            script_dir = get_lib_parent_dir()
            self.tk.call('wm', 'iconphoto', self.newWindow, ImageTk.PhotoImage(
                file=os.path.join(os.path.join(
                    script_dir,
                    'utils',
                    'resources',
                    'Assets',
                    'VAILabsIcon.ico'))))
            # self.newWindow.geometry("350x400")

            frame1 = tk.Frame(self.newWindow)
            frame4 = tk.Frame(self.newWindow)

            # Print settings
            tk.Label(frame1,
                     text="Please indicate your desired options for the "+self.plugin[self.m].get()+" plugin.", anchor=tk.N, justify=tk.LEFT).pack(expand=True)

            style = ttk.Style()
            style.configure(
                "Treeview", background='white', foreground='white',
                rowheight=25, fieldbackground='white',
                # font=self.controller.pages_font
                )
            style.configure("Treeview.Heading", 
            # font=self.controller.pages_font
            )
            style.map('Treeview', background=[('selected', 'grey')])

            frame2 = tk.Frame(self.newWindow, bg='green')
            self.r = 1
            self.create_treeView(frame2)
            self.fill_treeview(self.req_settings, self.opt_settings)
            frame2.grid(column=0, row=1, sticky="nswe", pady=10, padx=10)
    
            frame2.grid_rowconfigure(tuple(range(self.r)), weight=1)
            frame2.grid_columnconfigure(tuple(range(2)), weight=1)

            self.finishButton = tk.Button(
                frame4, text='Finish', command=self.removewindow)
            self.finishButton.grid(
                column=1, row=0, sticky="es", pady=(0, 10), padx=(0, 10))
            self.finishButton.bind(
                "<Return>", lambda event: self.removewindow())
            self.newWindow.protocol('WM_DELETE_WINDOW', self.removewindow)

            frame1.grid(column=0, row=0, sticky="ew")
            frame4.grid(column=0, row=2, sticky="se")
            self.newWindow.grid_rowconfigure(1, weight=2)
            self.newWindow.grid_columnconfigure(0, weight=1)

    def create_treeView(self, tree_frame):
        """ Function to create a new tree view in the given frame

        Parameters
        ----------
        tree_frame : tk.Frame
                    frame where the tree view will be included
        key : str
              key for the tree dictionary
        """

        tree_scrollx = tk.Scrollbar(tree_frame, orient='horizontal')
        tree_scrollx.pack(side=tk.BOTTOM, fill=tk.X)
        tree_scrolly = tk.Scrollbar(tree_frame)
        tree_scrolly.pack(side=tk.RIGHT, fill=tk.Y)

        self.tree = ttk.Treeview(tree_frame,
                                yscrollcommand=tree_scrolly.set,
                                xscrollcommand=tree_scrollx.set)
        self.tree.pack(fill='both', expand=True)

        tree_scrollx.config(command=self.tree.xview)
        tree_scrolly.config(command=self.tree.yview)

        columns_names = ['Name', 'Type', 'Value']
        self.tree['columns'] = columns_names

        # Format columns
        self.tree.column("#0", width=20,
                        minwidth=0, stretch=tk.NO)
        for n, cl in enumerate(columns_names):
            self.tree.column(
                cl, width=int(self.controller.pages_font.measure(str(cl)))+20,
                minwidth=50, anchor=tk.CENTER)
        # Headings
        for cl in columns_names:
            self.tree.heading(cl, text=cl, anchor=tk.CENTER)
        self.tree.tag_configure('req', foreground='black',
                                background='#9fc5e8')
        self.tree.tag_configure('opt', foreground='black',
                                background='#cfe2f3')
        self.tree.tag_configure('type', foreground='black',
                                background='#E8E8E8')
        self.tree.tag_configure('func', foreground='black',
                                background='#DFDFDF')
        # Define double-click on row action
        self.tree.bind("<Double-1>", self.OnDoubleClick)

    def OnDoubleClick(self, event):
        """ Executed when a row of the treeview is double clicked.
        Opens an entry box to edit a cell. """

        # ii = self.notebook.index(self.notebook.select())
        self.treerow = self.tree.identify_row(event.y)
        self.treecol = self.tree.identify_column(event.x)
        tags = self.tree.item(self.treerow)["tags"]
        if len(tags) > 0 and tags[0] in ['opt', 'req']:
            # get column position info
            x, y, width, height = self.tree.bbox(self.treerow, self.treecol)

            # y-axis offset
            pady = height // 2
            # pady = 0

            if hasattr(self, 'entry'):
                self.entry.destroy()

            self.entry = tk.Entry(self.tree, justify='center')

            if int(self.treecol[1:]) > 0:
                value = self.tree.item(self.treerow)['values'][int(str(self.treecol[1:]))-1] 
                value = str(value) if str(value) not in ['default', 'Choose X or Y'] else ''
                self.entry.insert(0, value)
                # self.entry['selectbackground'] = '#123456'
                self.entry['exportselection'] = False

                self.entry.focus_force()
                self.entry.bind("<Return>", self.on_return)
                self.entry.bind("<Escape>", lambda *ignore: self.entry.destroy())

                self.entry.place(x=x,
                                y=y + pady,
                                anchor=tk.W, width=width)
    
    def on_return(self, event):
        """ Executed when the entry is edited and pressed enter.
        Saves the edited value"""

        val = self.tree.item(self.treerow)['values']
        val[int(self.treecol[1:])-1] = self.entry.get()
        if self.entry.get() != '':
            self.tree.item(self.treerow, values=tuple([val[0], val[1], self.entry.get()]))
        elif val[2] == '':
            self.tree.item(self.treerow, values=tuple([val[0], val[1], 'default']))
        else:
            self.tree.item(self.treerow, values=val)
        self.entry.destroy()
        self.saved = False

    def fill_treeview(self, req_settings, opt_settings, parent = ''):
        """ Adds an entry for each setting. Displays it in the specified row.
        :param req_settings: dict type of plugin required setting options
        :param opt_settings: dict type of plugin optional setting options
        :param parent: string type of parent name
        """
        self.tree.insert(parent=parent, index='end', iid=parent+'_req', text='',
            values=tuple(['Required settings', '', '']), tags=('type',))
        self.r+=1
        for arg, val in req_settings.items():
            if arg == 'Data':
                self.tree.insert(parent=parent+'_req', index='end', iid=str(self.r), text='',
                    values=tuple([arg, val, 'Choose X or Y']), tags=('req',))
            else:
                self.tree.insert(parent=parent+'_req', index='end', iid=str(self.r), text='',
                                    values=tuple([arg, val, '']), tags=('req',))
            self.r+=1
        self.tree.insert(parent=parent, index='end', iid=parent+'_opt', text='',
            values=tuple(['Optional settings', '', '']), tags=('type',))
        self.r+=1
        for arg, val in opt_settings.items():
            self.tree.insert(parent=parent+'_opt', index='end', iid=str(self.r), text='',
                                 values=tuple([arg, val, 'default']), tags=('opt',))
            self.r+=1

    def removewindow(self):
        """ Stores settings options and closes window """
        self.req_settings.pop("Data", None)
        children = self.get_all_children()
        for child in children:
            tag = self.tree.item(child)["tags"][0]            
            if tag in ['req', 'opt']:
                val = self.tree.item(child)["values"]
                self.settingOptions(tag, val) 
        self.newWindow.destroy()
        self.newWindow = None
        self.focus()
    
    def get_all_children(self, item=""):
        """ Iterates over the treeview to get all childer """
        children = self.tree.get_children(item)
        for child in children:
            children += self.get_all_children(child)
        return children

    def settingOptions(self, tag, val):
        """ Identifies how the data should be stored """
        if val[0] == 'Data':
            if val[2] == 'Choose X or Y' or len(val[2]) == 0:
                self.updateSettings(tag, val[0], 'X')
            else:
                self.updateSettings(tag, val[0], val[2])
        else:
            if val[2] == 'default' or len(str(val[2])) == 0:
                self.updateSettings(tag, val[0])
            else:
                self.updateSettings(tag, val[0], val[2])

    def updateSettings(self, tag, key, value = None):
        """ Return the selected settings 

        Parameters
        ----------
        tag : str
              tag for the settings
        """
        if tag == 'req':
            if value is not None or self.req_settings[key] != value:
                self.req_settings[key] = value
            else:
                self.req_settings.pop(key, None)
        elif tag == 'opt':
            if value is not None or self.opt_settings[key] != value:
                self.opt_settings[key] = value
            else:
                self.opt_settings.pop(key, None)

    def on_return_entry(self, r):
        """ Changes focus to the next available entry. When no more, focuses 
        on the finish button.
        : param r: int type of entry id.
        """
        if r < len(self.entry):
            self.entry[r].focus()
        else:
            self.finishButton.focus()

    def CreateToolTip(self, widget, text):
        """ Calls ToolTip to create a window with a widget description. """
        toolTip = ToolTip(widget)

        def enter(event):
            toolTip.showtip(text)

        def leave(event):
            toolTip.hidetip()
        widget.bind('<Enter>', enter)
        widget.bind('<Leave>', leave)
        
    def module_out(self, name, iid):
        """ Updates the output DataFrame.

        :param name: name of the model
        :type name: str
        """
        name_list = list(self.out_data.columns)
        m_num = [n for n in name_list if (n.split('-')[0] == name)]
        if len(m_num) > 0:
            name_list.append(name + '-' + str(len(m_num)))
        else:
            name_list.append(name)
        self.canvas.itemconfig('t'+str(iid), text = name_list[-1])
        values = self.out_data.values
        values = np.vstack((
            np.hstack((values, np.zeros((values.shape[0], 1)))),
            np.zeros((1, values.shape[0]+1))))
        self.out_data = pd.DataFrame(values,
                                     columns=name_list,
                                     index=name_list)
        self.module_names.append(name_list[-1])

    def add_module(self,boxName: str,x: float,y:float,ini = False,out = False, iid = None):
        """ Creates a rectangular module with the corresponding text inside.

        :param boxName: name of the model
        :type boxName: str
        :param x: float type of module x coordinate
        :param y: float type of module y coordinate
        :param ini: bool type of whether the module corresponds to initialiser.
        :param out: bool type of whether the module corresponds to output.
        """
        iid = self.modules if iid is None else iid
        if not ini and not out:
            tag = ('o'+str(iid),)
        else: #Make initialisation and output unmoveable
            tag = ('n0',)
        text_w = self.controller.pages_font.measure(boxName+'-00') + 10
        self.canvas.create_rectangle(
            x - text_w/2 , 
            y - self.h/2, 
            x + text_w/2, 
            y + self.h/2, 
            tags = tag + ('p'+str(iid),), 
            fill = self.bg, 
            width = 3,
            # activefill = '#dbaa21'
        )
        self.canvas.create_text(
            x, 
            y, 
            font = self.controller.pages_font, 
            text = boxName, 
            tags = tag + ('t'+str(iid),), 
            fill = '#d0d4d9', 
            justify = tk.CENTER)
        
        if not out:
            self.canvas.create_oval(
                x - self.cr, 
                y + self.h/2 - self.cr, 
                x + self.cr, 
                y + self.h/2 + self.cr, 
                width = 2, 
                fill = 'black', 
                tags = tag + ('d'+str(iid),))
            
        if not ini:
            self.canvas.create_oval(
                x - self.cr, 
                y - self.h/2 - self.cr, 
                x + self.cr, 
                y - self.h/2 + self.cr, 
                width = 2, 
                fill = 'black', 
                tags = tag + ('u'+str(iid),))
        
        if not out and not ini:
            self.canvas.create_oval(
                x - text_w/2 - self.cr, 
                y - self.cr, 
                x - text_w/2 + self.cr, 
                y + self.cr, 
                width = 2, 
                fill = 'black', 
                tags = tag + ('l'+str(iid),))
        
            self.canvas.create_oval(
                x + text_w/2 - self.cr, 
                y - self.cr, 
                x + text_w/2 + self.cr, 
                y + self.cr, 
                width = 2, 
                fill = 'black', 
                tags = tag + ('r'+str(iid),))
            
        self.canvas_startxy.append((x, y))
        self.connections[iid] = {}
        self.module_out(boxName, iid)
        self.module_list.append(boxName)
        self.modules += 1

    def upload(self):
        """ Opens the XML file that was previously uploaded and places the 
        modules, loops and connections in the canvas."""

        filename = self.controller.output["xml_filename"]

        self.reset()

        self.s = XML_handler()
        self.s.load_XML(filename)
        # self.s._print_pretty(self.s.loaded_modules)
        modules = self.s.loaded_modules
        modout = modules['Output']
        # They are generated when resetting
        del modules['Initialiser'], modules['Output']
        self.disp_mod = ['Initialiser', 'Output']
        self.id_mod = [0, 1]

        # Place the modules
        self.place_modules(modules)
        self.draw_connection(modules)
        connect = list(modout['coordinates'][2].keys())
        for p, parent in enumerate(modout['parents']):
            parent_id = self.id_mod[np.where(
                np.array(self.disp_mod) == parent)[0][0]]
            out, ins = modout['coordinates'][2][connect[p]].split('-')
            xout, yout, _, _ = self.canvas.coords(out[0]+str(parent_id))
            xins, yins, _, _ = self.canvas.coords(ins[0]+str(1))
            self.canvas.create_line(
                xout + self.cr,
                yout + self.cr,
                xins + self.cr,
                yins + self.cr,
                fill = "red",  width = 2,
                arrow = tk.LAST, arrowshape = (12,10,5), 
                tags=('o'+str(parent_id),
                      'o'+str(1), modout['coordinates'][2][connect[p]]))
            self.out_data.iloc[int(parent_id)][1] = 1
            self.connections[1][
                int(parent_id)] = out[0]+str(parent_id) + '-' + ins[0]+str(1)
        self.m: int = self.id_mod[2]
        x0, y0, x1, y1 = self.canvas.coords('p'+str(self.m))

        # Configure frame for scrollbar
        self.frame_canvas.configure(
            yscrollcommand=self.tree_scrolly.set, xscrollcommand=self.tree_scrollx.set)
        self.frame_canvas.bind('<Configure>', lambda e: self.frame_canvas.configure(
            scrollregion=self.frame_canvas.bbox("all")))
        self.framep = tk.Frame(self.frame_canvas, bg=self.bg)
        self.frame_canvas.create_window((0, 0), window=self.framep)

        self.set_mousewheel(
            self.frame_canvas, lambda e: self.frame_canvas.yview_scroll(-1*(e.delta//120), "units"))

        self.select(x0, y0)

        self.framex.grid(column=1, row=1, sticky="nswe",
                         rowspan=4, pady=(0, 10))
        self.framex.grid_columnconfigure(tuple(range(2)), weight=1)
        self.frame_canvas.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

    def set_mousewheel(self, widget, command):
        """Activate / deactivate mousewheel scrolling when 
        cursor is over / not over the widget respectively."""
        widget.bind("<Enter>", lambda _: widget.bind_all(
            '<MouseWheel>', command))
        widget.bind("<Leave>", lambda _: widget.unbind_all('<MouseWheel>'))
            
    def place_modules(self, modules):
        # Place the modules
        for key in [key for key, val in modules.items() if type(val) == dict]:
            if modules[key]['class'] == 'loop':
                self.l += 1
                # Extracts numbers from string
                l = int(
                    ''.join(map(str, list(filter(str.isdigit, modules[key]['name'])))))
                x0, y0, x1, y1 = modules[key]['coordinates']
                self.canvas.create_rectangle(x0, y0,
                                             x1, y1,
                                             outline='#4ff07a',
                                             tag='loop-'+str(l))  # type: ignore
                text_w = self.controller.pages_font.measure(
                    modules[key]['type']) + 20
                self.canvas.create_text(
                    x0 + text_w/2,
                    y0 + 20,
                    font=self.controller.pages_font,
                    text=modules[key]['type'],
                    tags=('loop-'+str(l), 'type'+'-'+str(l)),
                    justify=tk.CENTER)
                text_w = self.controller.pages_font.measure(
                    modules[key]['condition']) + 20
                self.canvas.create_text(
                    x1 - text_w/2,
                    y0 + 20,
                    font=self.controller.pages_font,
                    text=modules[key]['condition'],
                    tags=('loop-'+str(l), 'condition'+'-'+str(l)),
                    justify=tk.CENTER)
                self.loops.append({'type': modules[key]['type'],
                                   'condition': modules[key]['condition'],
                                   'mod': [],
                                   'coord': (x0, y0,
                                             x1, y1)})
                self.place_modules(modules[key])
            elif key not in self.disp_mod:
                # Display module
                self.add_module(key,
                                modules[key]['coordinates'][0][0],
                                modules[key]['coordinates'][0][1], 
                                iid = modules[key]['coordinates'][1])
                self.module_list[-1] = modules[key]['module_type']
                self.id_mod.append(modules[key]['coordinates'][1])
                self.disp_mod.append(key)
        self.module_list = [x for _, x in sorted(zip(self.id_mod, self.module_list))]
        self.canvas_startxy = [x for _, x in sorted(zip(self.id_mod, self.canvas_startxy))]

    def draw_connection(self, modules):
        for key in [key for key, val in modules.items() if type(val) == dict]:
            if modules[key]['class'] == 'loop':
                self.draw_connection(modules[key])
            else:
                # Connect modules
                connect = list(modules[key]['coordinates'][2].keys())
                for p, parent in enumerate(modules[key]['parents']):
                    if not (parent[:4] == 'loop'):
                        # parent_id = modules[parent]['coordinates'][1]
                        parent_id = self.id_mod[np.where(np.array(self.disp_mod) == parent)[0][0]]
                        out, ins = modules[key]['coordinates'][2][connect[p]].split('-')
                        xout, yout, _ , _ = self.canvas.coords(out)
                        xins, yins, _, _ = self.canvas.coords(ins)
                        self.canvas.create_line(
                                    xout + self.cr,
                                    yout + self.cr,
                                    xins + self.cr,
                                    yins + self.cr,
                                    fill = "red",  width = 2,
                                    arrow = tk.LAST, arrowshape = (12,10,5), 
                                    tags = ('o'+str(parent_id),
                                          # 'o'+str(self.id_mod[-1]), 
                                          modules[key]['coordinates'][2][connect[p]]))
                        self.out_data.iloc[int(parent_id)][int(ins[1:])] = 1
                        self.connections[int(ins[1:])][
                            int(parent_id)] = out[0]+str(out[1:]) + '-' + ins[0]+str(ins[1:])
                    else:
                        self.loops[-1]['mod'].append(key)

    def reset(self):
        """ Resets the canvas and the stored information."""
        self.canvas.delete(tk.ALL)  # Reset canvas

        if hasattr(self, 'newWindow') and (self.newWindow != None):
            self.newWindow.destroy()

        self.canvas_startxy = []
        self.out_data = pd.DataFrame()
        self.connections = {}
        self.modules = 0
        self.module_list = []
        self.module_names = []

        self.add_module('Initialiser', self.width/2, self.h, ini=True)
        self.add_module('Output', self.width/2, self.height - self.h, out=True)

        self.draw = False
        self.loops = []
        self.drawLoop = False
        self.l = 0
        self.id_done = [0, 1]
        self.plugin = {}
        for widget in self.allWeHearIs:
            widget.grid_remove()
        self.allWeHearIs = []
        self.my_label.config(text='')

    def check_quit(self):
        """Checks if all plugins are specified to reset and go back to the startpage"""
        if not len(self.id_done) == len(self.id_mod):
            response = messagebox.askokcancel(
                "Exit",
                "There are some unspecified plugins. Are you sure you want to leave?")
            if response:
                self.reset()
                self.canvas.delete(tk.ALL)
                self.frame_canvas.delete(tk.ALL)
                self.saved = True
                self.s.write_to_XML()
                self.controller.Plugin.set(True)
                self.controller._show_frame("MainPage")
        # TODO: Check if loaded
        elif len(self.s.loaded_modules) == 0:
            self.controller._show_frame("MainPage")
            self.controller.Plugin.set(False)
        else:
            self.reset()
            self.canvas.delete(tk.ALL)
            self.frame_canvas.delete(tk.ALL)
            self.s.write_to_XML()
            self.controller.Plugin.set(True)
            self.controller._show_frame("MainPage")


class ToolTip(object):
    """ Defines a text window associated to a widget with a description. """

    def __init__(self, widget):
        self.widget = widget
        self.tipwindow = None
        self.id = None
        self.x = self.y = 0

    def showtip(self, text):
        "Display text in tooltip window"
        self.text = text
        if self.tipwindow or not self.text:
            return
        x, y, cx, cy = self.widget.bbox("insert")
        x = x + self.widget.winfo_rootx() + 60
        y = y + cy + self.widget.winfo_rooty() + 50
        self.tipwindow = tw = tk.Toplevel(self.widget)
        tw.wm_overrideredirect(1)
        tw.wm_geometry("+%d+%d" % (x, y))
        label = tk.Label(tw, text=self.text, justify=tk.LEFT,
                         background="#d0d4d9", relief=tk.SOLID, borderwidth=1,
                         font=("Helvetica", "8", "normal"))
        label.pack(ipadx=1)

    def hidetip(self):
        tw = self.tipwindow
        self.tipwindow = None
        if tw:
            tw.destroy()


class EntryWithPlaceholder(tk.Entry):
    """ Defines an entry with a placeholder text displayed if not focused on."""

    def __init__(self, master=None, placeholder="PLACEHOLDER", color='grey'):
        super().__init__(master)

        self.placeholder = placeholder
        self.placeholder_color = color
        self.default_fg_color = self['fg']

        self.bind("<FocusIn>", self.foc_in)
        self.bind("<FocusOut>", self.foc_out)

        self.put_placeholder()

    def put_placeholder(self):
        self.insert(0, self.placeholder)
        self['fg'] = self.placeholder_color

    def foc_in(self, *args):
        if self['fg'] == self.placeholder_color:
            self.delete('0', 'end')
            self['fg'] = self.default_fg_color

    def foc_out(self, *args):
        if not self.get():
            self.put_placeholder()



from vai_lab.Data.xml_handler import XML_handler
from vai_lab._types import DictT

import os
import numpy as np
import pandas as pd
from typing import List, Literal,Tuple

from PIL import Image, ImageTk
import tkinter as tk
from tkinter import messagebox
from tkinter.filedialog import asksaveasfile, askopenfile, askopenfilename
from vai_lab._import_helper import get_lib_parent_dir


_PLUGIN_READABLE_NAMES = {"aid_canvas": "default",
                          "aidCanvas": "alias",
                          "aid": "alias",
                          "AID": "alias"}               # type:ignore
_PLUGIN_MODULE_OPTIONS = {"layer_priority": 2,
                          "required_children": None}    # type:ignore
_PLUGIN_REQUIRED_SETTINGS = {}                          # type:ignore
_PLUGIN_OPTIONAL_SETTINGS = {}                          # type:ignore


class aidCanvas(tk.Frame):
    """Canvas for graphical specification of pipeline modules"""

    def __init__(self, parent, controller, config: dict):
        " Here we define the main frame displayed upon opening the program."
        " This leads to the different methods to provide feedback."

        super().__init__(parent, bg=parent['bg'])
        self.bg = parent['bg']
        self.controller = controller

        script_dir = get_lib_parent_dir()
        self.tk.call('wm', 'iconphoto', self.controller._w, ImageTk.PhotoImage(
            file=os.path.join(os.path.join(
                script_dir,
                'utils',
                'resources',
                'Assets',
                'VAILabsIcon.ico'))))
        self.my_img1 = ImageTk.PhotoImage(Image.open(os.path.join(
            script_dir,
            'utils',
            'resources',
            'Assets',
            'VAILabs.png')).resize((200, 200)))

        # self.grid_rowconfigure(tuple(range(2)), weight=1)
        self.grid_rowconfigure(1, weight=1)
        self.grid_columnconfigure(0, weight=1)

        frame1 = tk.Frame(self, bg=self.bg)
        frame2 = tk.Frame(self, bg=self.bg)
        frame3 = tk.Frame(self, bg=self.bg)
        frame4 = tk.Frame(self, bg=self.bg)

        self.my_label = tk.Label(frame2, image=self.my_img1, bg=parent['bg'], anchor = 'center')
        self.my_label.grid(column=5, row=0, padx=(10, 10), pady=(10, 10), sticky="nwse")

        # Create canvas
        self.width, self.height = 700, 700
        self.canvas = tk.Canvas(frame1, width=self.width,
                                height=self.height, background="white")
        # self.canvas = ResizingCanvas(frame1, width=self.width,
        #     height=self.height, background="white")
        self.canvas.pack(fill=tk.BOTH, expand=True, padx=(10, 0), pady=10)

        self.canvas_startxy: List[Tuple] = []
        self.out_data = pd.DataFrame()
        self.connections: DictT = {}
        self.modules = 0
        self.module_list: List[str] = []
        self.module_names: List[str] = []

        # Create module
        self.w, self.h = 100, 50
        self.cr = 4

        # Initialiser module
        self.add_module('Initialiser', self.width/2, self.h, ini=True)
        self.add_module('Output', self.width/2, self.height - self.h, out=True)

        self.draw = False
        self.canvas.bind("<B1-Motion>", self.on_drag)
        self.canvas.bind('<Button-1>', self.select)
        self.l = 0  # number of loops
        self.drawLoop = False
        self.loops: List[DictT] = []

        # self.canvas.bind("<ButtonPress-1>", self.on_button_press)
        self.canvas.bind("<ButtonRelease-1>", self.on_button_release)

        # for m, module in enumerate(modules):
        tk.Button(
            frame4, text='Data processing', fg='white', bg=parent['bg'],
            height=3, width=25, font=self.controller.pages_font,
            command=lambda: self.add_module('Data Processing',
                                            self.width/2,
                                            self.height/2)
        ).grid(column=5, row=1, padx=(10, 10), sticky="news")
        tk.Button(
            frame4, text='Modelling', fg='white', bg=parent['bg'],
            height=3, width=25, font=self.controller.pages_font,
            command=lambda: self.add_module('Modelling',
                                            self.width/2,
                                            self.height/2)
        ).grid(column=5, row=2, padx=(10, 10), sticky="news")
        tk.Button(
            frame4, text='Decision making', fg='white', bg=parent['bg'],
            height=3, width=25, font=self.controller.pages_font,
            command=lambda: self.add_module('Decision Making',
                                            self.width/2,
                                            self.height/2)
        ).grid(column=5, row=3, padx=(10, 10), sticky="news")
        tk.Button(
            frame4, text='User Interaction', fg='white', bg=parent['bg'],
            height=3, width=25, font=self.controller.pages_font,
            command=lambda: self.add_module('User Interaction',
                                            self.width/2,
                                            self.height/2)
        ).grid(column=5, row=4, padx=(10, 10), sticky="news")
        tk.Button(
            frame4, text = 'Input data', fg = 'white', bg = parent['bg'],
            height = 3, width = 25, font = self.controller.pages_font,
            command = lambda: self.add_module('Input Data', 
                                              self.width/2, 
                                              self.height/2)
            ).grid(column = 5, row = 5, padx=(10,10), sticky="news")        
        tk.Button(
            frame4, text = 'Data Storage', fg = 'white', bg = parent['bg'],
            height = 3, width = 25, font = self.controller.pages_font,
            command = lambda: self.add_module('Data Storage', 
                                              self.width/2, 
                                              self.height/2)
            ).grid(column = 5, row = 6, padx=(10,10), sticky="news")
        tk.Button(
            frame4, text = 'Delete selection', fg = 'white', bg = parent['bg'],
            height = 3, width = 25, font = self.controller.pages_font,
            command = self.delete_sel).grid(column = 5, row = 7, sticky="news"
                                            , padx=(10,10), pady=(0,10))
        
        ww = int(self.width/40)
        tk.Button(
            frame3, text='Upload', fg='white', bg=parent['bg'],
            height=3, width=ww, font=self.controller.pages_font,
            command=self.upload).grid(column=0, row=10, sticky="news",
                                      padx=(10, 0), pady=(0, 10))
        tk.Button(
            frame3, text='Save', fg='white', bg=parent['bg'],
            height=3, width=ww, font=self.controller.pages_font,
            command=self.save_file).grid(column=1, row=10, sticky="news", pady=(0, 10))
        tk.Button(
            frame3, text='Reset', fg='white', bg=parent['bg'],
            height=3, width=ww, font=self.controller.pages_font,
            command=self.reset).grid(column=2, row=10, sticky="news", pady=(0, 10))
        tk.Button(
            frame3, text='Done', fg='white', bg=parent['bg'],
            height=3, width=ww, font=self.controller.pages_font,
            command=self.check_quit).grid(column=3, row=10, sticky="news", pady=(0, 10))

        self.save_path = ''
        self.saved = True

        frame1.grid(column=0, row=0, rowspan=2, sticky="nsew")
        frame2.grid(column=1, row=0, sticky="nwse")
        frame3.grid(column=0, row=2, sticky="swe")
        frame4.grid(column=1, row=1, sticky="nse")
        
        frame3.grid_columnconfigure(tuple(range(4)), weight=1)
        frame4.grid_rowconfigure(7, weight=1)

    def class_list(self, value):
        """ Temporary fix """
        return value

    def on_return_display(self, event):
        """ Defines the type of loop and condition for the indicated loop
        """
        if self.loopDisp:
            condition = self.entry1.get()
            self.entry1.destroy()
            if hasattr(self, 'entry2'):
                self.entry2.focus()
            key = 'type'
            text_w = self.controller.pages_font.measure(condition) + 20
            x = self.start_x + text_w/2
        else:
            condition = self.entry2.get()
            self.entry2.destroy()
            key = 'condition'
            text_w = self.controller.pages_font.measure(condition) + 20
            x = self.curX - text_w/2
        self.loops[-1][key] = condition
        self.canvas.create_text(
            x,
            self.start_y + 20,
            font=self.controller.pages_font,
            text=condition,
            tags=('loop-'+str(self.l-1), key+'-'+str(self.l-1)),
            justify=tk.CENTER)
        self.canvas.tag_bind(key+'-'+str(self.l-1),
                             "<Double-1>", self.OnDoubleClick)
        self.loopDisp = not self.loopDisp

    def on_button_release(self, event):
        """ Finishes drawing the rectangle for loop definition. 
            Once the rectangle is drawn, it stores the information of the 
            modules inside the loop.
        """
        if self.drawLoop:
            self.drawLoop = False

            if (abs(self.start_x - self.curX) > 10) and (
                    abs(self.start_y - self.curY) > 10):
                self.l += 1
                # Identify modules in area
                loopMod = self.canvas.find_enclosed(self.start_x, self.start_y,
                                                    self.curX, self.curY)
                self.loops.append({'type': None,
                                   'condition': None,
                                   'mod': [],
                                   'coord': (self.start_x, self.start_y,
                                             self.curX, self.curY)})
                for mod in loopMod:
                    aux = self.canvas.itemcget(mod, 'tags').split(' ')
                    if aux[0][:4] != 'loop':
                        if len(aux) == 2 and len(aux[1]) > 1 and aux[1][0] == 't':
                            self.loops[-1]['mod'].append(self.canvas.itemcget(
                                mod, 'text'))
                
                # Ask for loop definition and condition and display               
                self.entry1 = tk.Entry(self.canvas, justify='center', 
                                      font = self.controller.pages_font)
                text = 'For/While'
                self.loopDisp = True
                self.entry1.insert(0, text)
                # self.entry['selectbackground'] = '#d0d4d9'
                self.entry1['exportselection'] = False

                self.entry1.focus()
                self.entry1.bind("<Return>", self.on_return_display)
                self.entry1.bind(
                    "<Escape>", lambda *ignore: self.entry1.destroy())

                self.entry1.place(
                    x=self.start_x + 10,
                    y=self.start_y + 20,
                    anchor=tk.W,
                    width=self.controller.pages_font.measure(text)+10)
                text = 'Condition'
                self.entry2 = tk.Entry(self.canvas, justify='center',
                                       font=self.controller.pages_font)
                self.entry2.insert(0, text)
                self.entry2['exportselection'] = False

                self.entry2.bind("<Return>", self.on_return_display)
                self.entry2.bind(
                    "<Escape>", lambda *ignore: self.entry.destroy())

                self.entry2.place(
                    x=self.curX-self.controller.pages_font.measure(text)-20,
                    y=self.start_y + 20,
                    anchor=tk.W,
                    width=self.controller.pages_font.measure(text)+10)

                self.saved = False
            else:
                self.canvas.delete('loop-'+str(self.l))
        else:
            self.updateLoops()

    def updateLoops(self):
        """ Checks if, after some movement, a module is included in an existing
            loop
        """
        for l, loop in enumerate(self.loops):
            coord = loop['coord']
            loopMod = self.canvas.find_enclosed(coord[0], coord[1],
                                                coord[2], coord[3])
            self.loops[l]['mod'] = []
            for mod in loopMod:
                aux = self.canvas.itemcget(mod, 'tags').split(' ')[-1]
                if len(aux) == 2 and aux[0] == 't':
                    self.loops[l]['mod'].append(self.canvas.itemcget(
                        mod, 'text'))
        # print('Updated loop info:', self.loops)
        
    def select(self, event):
        """ Selects the module at the mouse location. """
        self.selected = self.canvas.find_overlapping(
            event.x-5, event.y-5, event.x+5, event.y+5)
        if self.selected:
            if len(self.selected) > 2:
                self.canvas.selected = self.selected[-2]
            else:
                self.canvas.selected = self.selected[-1]
            # self.canvas_startxy = (event.x, event.y)
        else:
            self.canvas.selected = None
            # save mouse drag start position
            self.start_x = self.canvas.canvasx(event.x)
            self.start_y = self.canvas.canvasy(event.y)
            self.curX = self.start_x
            self.curY = self.start_y
            self.selected = self.canvas.find_overlapping(
                event.x-5, event.y-5, event.x+5, event.y+5)

            # create rectangle if not yet exist
            if not self.selected:
                self.rect = self.canvas.create_rectangle(event.x, event.y,
                                                         event.x, event.y,
                                                         outline='#4ff07a',
                                                         tag='loop-'+str(self.l))
                # ,fill = '#b3ffc7')
                self.drawLoop = True
                if self.l == 0:
                    self.canvas.tag_lower('loop-'+str(self.l))
                else:
                    self.canvas.tag_raise(
                        'loop-'+str(self.l), 'loop-'+str(self.l-1))

        if len(self.canvas.gettags("current")) > 0:
            if any('loop' in value for value in self.canvas.gettags("current")):
                self.isLoop = True
                self.m = [int(el.split('-')[-1]) for el in self.canvas.gettags("current") if 'loop' in el][0]
            else:
                self.isLoop = False
                self.canvas.gettags("current")
                #Preference to objects with text
                if any(el[0] == 't' for el in self.canvas.gettags("current")):
                    self.m = [int(el[1:]) for el in self.canvas.gettags("current") if el[0] == 't'][0]
                else:
                    self.m = [int(el[1:]) for el in self.canvas.gettags("current") if el[1:].isdigit()][0]

    def on_drag(self, event):
        """ Uses the mouse location to move the module and its text. 
        At the same time, it looks up if there are any connection to this
        module and subsequently moves the connection."""

        if self.drawLoop:
            self.curX = self.canvas.canvasx(event.x)
            self.curY = self.canvas.canvasy(event.y)

            # w, h = self.canvas.winfo_width(), self.canvas.winfo_height()
            # if event.x > 0.9*w:
            #     self.canvas.xview_scroll(1, 'units')
            # elif event.x < 0.1*w:
            #     self.canvas.xview_scroll(-1, 'units')
            # if event.y > 0.9*h:
            #     self.canvas.yview_scroll(1, 'units')
            # elif event.y < 0.1*h:
            #     self.canvas.yview_scroll(-1, 'units')

            # expand rectangle as you drag the mouse
            self.canvas.coords('loop-'+str(self.l), self.start_x,
                               self.start_y, self.curX, self.curY)
        else:
            self.select(event)
            sel_obj = [int(el[1:]) for el in self.canvas.gettags("current") if el[0] == 't']
            if len(sel_obj) > 0:
                self.m = sel_obj[0]
                
                dx = event.x - self.canvas_startxy[self.m][0]
                dy = event.y - self.canvas_startxy[self.m][1]
                # if self.isLoop:
                    # self.canvas.move('loop-'+str(self.m), dx, dy) 
                # else:
                # module
                self.canvas.move('o'+str(self.m), dx, dy)
                # self.updateLoops()
                # Connections
                if any(self.out_data.iloc[self.m].values) or any(
                        self.out_data[self.out_data.columns[self.m]].values):
                    out = np.arange(self.out_data.values.shape[0])[
                        self.out_data.values[self.m, :]==1]
                    for o in out:
                        xycoord_i = self.canvas.coords(
                            self.connections[o][self.m].split('-')[0])
                        xycoord_o = self.canvas.coords(
                            self.connections[o][self.m].split('-')[1])
                        self.canvas.coords(self.connections[o][self.m], 
                                (xycoord_i[0] + self.cr, 
                                 xycoord_i[1] + self.cr, 
                                 xycoord_o[0] + self.cr, 
                                 xycoord_o[1] + self.cr))
                    inp = np.arange(self.out_data.values.shape[0])[
                        self.out_data.values[:, self.m]==1]
                    for i in inp:
                        xycoord_i = self.canvas.coords(
                            self.connections[self.m][i].split('-')[0])
                        xycoord_o = self.canvas.coords(
                            self.connections[self.m][i].split('-')[1])
                        self.canvas.coords(self.connections[self.m][i], 
                                (xycoord_i[0] + self.cr, 
                                 xycoord_i[1] + self.cr, 
                                 xycoord_o[0] + self.cr, 
                                 xycoord_o[1] + self.cr))
                # Update module location
                self.canvas_startxy[self.m] = (event.x, event.y)
    
    def module_out(self, name, iid):
        """ Updates the output DataFrame.

        :param name: name of the model
        :type name: str
        """
        name_list = list(self.out_data.columns)
        m_num = [n for n in name_list if (n.split('-')[0] == name)]
        if len(m_num) > 0:
            name_list.append(name + '-' + str(len(m_num)))
        else:
            name_list.append(name)
        self.canvas.itemconfig('t'+str(iid), text = name_list[-1])
        values = self.out_data.values
        values = np.vstack((
            np.hstack((values, np.zeros((values.shape[0], 1)))),
            np.zeros((1, values.shape[0]+1))))
        self.out_data = pd.DataFrame(values,
                                     columns=name_list,
                                     index=name_list)
        self.module_names.append(name_list[-1])

    def add_module(self, boxName, x, y, ini = False, out = False, iid = None):
        """ Creates a rectangular module with the corresponding text inside.

        :param boxName: name of the model
        :type boxName: str
        """
        iid = self.modules if iid is None else iid
        if not ini and not out:
            tag = ('o'+str(iid),)
        else: #Make initialisation and output unmoveable
            tag = ('n0',)
        text_w = self.controller.pages_font.measure(boxName+'-00') + 10
        self.canvas.create_rectangle(
            x - text_w/2 , 
            y - self.h/2, 
            x + text_w/2, 
            y + self.h/2, 
            tags = tag + ('p'+str(iid),), 
            fill = self.bg, width = 3,
            activefill = '#dbaa21')
        self.canvas.create_text(
            x, 
            y, 
            font = self.controller.pages_font, 
            text = boxName, 
            tags = tag + ('t'+str(iid),), 
            fill = '#d0d4d9', 
            justify = tk.CENTER)
        self.canvas.tag_bind('t'+str(iid), 
                             "<Double-1>", self.OnDoubleClick)
        if not out:
            self.canvas.create_oval(
                x - self.cr, 
                y + self.h/2 - self.cr, 
                x + self.cr, 
                y + self.h/2 + self.cr, 
                width = 2, 
                fill = 'black', 
                tags = tag + ('d'+str(iid),))
            self.canvas.tag_bind('d'+str(iid), 
                                 "<Button-1>", self.join_modules)

        if not ini:
            self.canvas.create_oval(
                x - self.cr, 
                y - self.h/2 - self.cr, 
                x + self.cr, 
                y - self.h/2 + self.cr, 
                width = 2, 
                fill = 'black', 
                tags = tag + ('u'+str(iid),))
            self.canvas.tag_bind('u'+str(iid), 
                                 "<Button-1>", self.join_modules)

        if not out and not ini:
            self.canvas.create_oval(
                x - text_w/2 - self.cr, 
                y - self.cr, 
                x - text_w/2 + self.cr, 
                y + self.cr, 
                width = 2, 
                fill = 'black', 
                tags = tag + ('l'+str(iid),))
            self.canvas.tag_bind('l'+str(iid), 
                                 "<Button-1>", self.join_modules)

            self.canvas.create_oval(
                x + text_w/2 - self.cr, 
                y - self.cr, 
                x + text_w/2 + self.cr, 
                y + self.cr, 
                width = 2, 
                fill = 'black', 
                tags = tag + ('r'+str(iid),))
            self.canvas.tag_bind('r'+str(iid), 
                                 "<Button-1>", self.join_modules)
        self.canvas_startxy.append((x,
                                    y))
        self.connections[iid] = {}
        self.module_out(boxName, iid)
        self.module_list.append(boxName)
        self.modules += 1
        self.saved = False

    def on_return_rename(self, event):
        """ This function renames a module to the input specified through 
        an entry widget. It then updates the corresponding rows and columns in
        the stored data."""

        moduleName = self.entry.get()
        if (moduleName in list(self.module_names)) and not(
                moduleName == list(self.module_names)[self.m]):
            messagebox.showwarning("Error", "This module already exists.")
        else:
            self.canvas.itemconfig('t'+str(self.m), text=moduleName)
            x0, y0, x1, y1 = self.canvas.coords('p'+str(self.m))
            text_w = self.controller.pages_font.measure(moduleName) + 20
            shift = (text_w-x1+x0)/2
            self.canvas.coords('p'+str(self.m),  # Resize rectangle
                               x0 - shift,
                               y0,
                               x1 + shift,
                               y1)
            x0, y0, x1, y1 = self.canvas.coords('l'+str(self.m))
            self.canvas.coords('l'+str(self.m),  # Move left circle
                               x0 - shift,
                               y0,
                               x1 - shift,
                               y1)
            x0, y0, x1, y1 = self.canvas.coords('r'+str(self.m))
            self.canvas.coords('r'+str(self.m),  # Move left circle
                               x0 + shift,
                               y0,
                               x1 + shift,
                               y1)
            self.entry.destroy()
            self.module_names[self.m] = moduleName
            self.saved = False

    def on_return_editLoop(self, event):
        """ This function renames loop conditions and updates the loop
        information."""

        newText = self.entry.get()
        conditions = self.canvas.itemcget(
            self.selected[0], 'tags').split(' ')[1].split('-')
        self.canvas.itemconfig(self.selected[0], text=newText)
        self.entry.destroy()
        self.loops[int(conditions[1])][conditions[0]] = newText
        self.saved = False

    def OnDoubleClick(self, event):
        """ Executed when text is double clicked.
        Opens an entry box to edit the module name and updates the display and
        the stored data. """

        self.selected = self.canvas.find_overlapping(
            event.x-5, event.y-5, event.x+5, event.y+5)

        # If not loop text
        if len(self.canvas.itemcget(
                self.selected[0], 'tags').split(' ')[-2].split('-')) < 2:

            x0, y0, x1, y1 = self.canvas.coords(
                self.canvas.gettags("current")[0])
            if hasattr(self, 'entry'):
                self.entry.destroy()

            entryText = self.canvas.itemcget('t'+str(self.m), 'text')

            self.entry = tk.Entry(self.canvas, justify='center',
                                  font=self.controller.pages_font)

            self.entry.insert(
                0, entryText)
            self.entry['selectbackground'] = '#d0d4d9'
            self.entry['exportselection'] = False

            self.entry.focus_force()
            self.entry.bind("<Return>", self.on_return_rename)
            self.entry.bind("<Escape>", lambda *ignore: self.entry.destroy())

            self.entry.place(x=x0,
                             y=y0 + (y1-y0)/2,
                             anchor=tk.W, width=x1 - x0)
        # If loop text
        else:
            x0, y0 = self.canvas.coords(self.selected[0])
            if hasattr(self, 'entry'):
                self.entry.destroy()

            # condition = self.canvas.itemcget(
            #             self.selected[0], 'tags').split(' ')[-2]#.split('-')[0]
            entryText = self.canvas.itemcget(self.selected[0], 'text')

            self.entry = tk.Entry(self.canvas, justify='center',
                                  font=self.controller.pages_font)
            self.entry.insert(0, entryText)
            # self.entry1['selectbackground'] = '#d0d4d9'
            self.entry['exportselection'] = False
            
            self.entry.focus()
            self.entry.bind("<Return>", self.on_return_editLoop)
            self.entry.bind("<Escape>", lambda *ignore: self.entry.destroy())
            text_w = self.controller.pages_font.measure(entryText) + 20
            self.entry.place(x=x0 - text_w/2,
                             y=y0,
                             anchor=tk.W,
                             width=text_w)

    def delete_sel(self):
        """ Deletes last selected module"""
        if self.canvas.selected:
            if self.isLoop:
                self.canvas.delete('loop-'+str(self.m))
                self.loops[self.m] = -1
            elif not(self.m == 0):
                self.canvas.delete('o'+str(self.m))

                col = self.out_data.columns[self.m]
                self.out_data[col] = 0
                self.out_data.loc[col] = 0
                self.saved = False
                self.module_names[self.m] = None

    def join_modules(self, event):
        """ Draws a connecting line between two connecting circles. """
        if self.draw:
            tags = self.canvas.gettags(self.canvas.find_overlapping(
                event.x-self.cr, event.y-self.cr,
                event.x+self.cr, event.y+self.cr)[-1])
            tag2 = tags[-2] if tags[-1] == "current" else tags[-1]
            if tag2[0] in ['d', 'l', 'u', 'r'] and int(
                    tag2[1:]) != int(self.tag[1:]):
                self.canvas.create_line(
                            self.canvas.linestartxy[0] + self.cr, 
                            self.canvas.linestartxy[1] + self.cr, 
                            self.canvas.coords(tag2)[0] + self.cr, 
                            self.canvas.coords(tag2)[1] + self.cr,
                            fill = "red",  width = 2,
                            arrow = tk.LAST, arrowshape = (12,10,5), 
                            tags = ('o'+str(int(self.tag[1:])), 
                                  'o'+str(int(tag2[1:])), self.tag + '-' + tag2))
                self.out_data.iloc[int(self.tag[1:])][int(tag2[1:])] = 1
                self.connections[int(tag2[1:])][
                    int(self.tag[1:])] = self.tag + '-' + tag2
            self.draw = False
            self.saved = False
        else:
            tags = self.canvas.gettags(self.canvas.find_overlapping(
                event.x-self.cr, event.y-self.cr,
                event.x+self.cr, event.y+self.cr)[-1])
            self.tag = tags[-2] if tags[-1] == "current" else tags[-1]
            self.canvas.linestartxy = self.canvas.coords(
                self.tag)  # (event.x, event.y)
            self.draw = True

    def save_file_as(self):

        self.save_path = asksaveasfile(mode='w')
        self.save_file()

    def save_file(self):

        if self.save_path == '':
            self.save_path = asksaveasfile(defaultextension='.xml', filetypes=[('XML file', '.xml'),
                                                                               ('All Files', '*.*')])
        # asksaveasfile return `None` if dialog closed with "cancel".
        if self.save_path is not None:
            # Transform input for output file
            data = self.out_data.copy()
            if np.sum(data.values[0, :]) == 0:
                messagebox.showwarning(
                    "Error", "Initialiser is not connected to any module.")
            elif np.sum(data.values[:, 1]) == 0:
                messagebox.showwarning(
                    "Error", "Output is not connected to any module.")
            idx = (np.sum(data.values, axis=0)
                   + np.sum(data.values, axis=1)) < 1
            col = np.array(data.columns)
            for c in np.arange(len(idx))[idx]:
                data = data.drop(columns=col[c], index=col[c])

            loop_modules = np.unique([v for a in self.loops for v in a['mod']])
            out_loops = np.zeros_like(self.loops)

            values = data.values.astype(bool)
            mn = np.array(self.module_names)
            mn_id = [m is not None for m in mn]
            mn = mn[mn_id]

            # to avoid numpy bug during elementwise comparison of lists 
            loop_modules.dtype = mn.dtype if len(loop_modules) == 0 else loop_modules.dtype
            
            self.controller.s.update_module_coords(self.module_list[0],[self.canvas_startxy[0], 0, self.connections[0]])
            self.controller.s.append_module_relationships(self.module_list[0],list(mn[values[:,0]]),list(mn[values[0,:]]))
            self.controller.s.filename = self.save_path.name
            for i, mnn in enumerate(mn_id):
                if (i > 1) and mnn:
                    xml_parent = None
                    parent_loops = []
                    if mn[i] in loop_modules:  # Model in any loop
                        for x, loop in enumerate(self.loops):
                            if mn[i] in loop['mod']:  # Model in this loop
                                if not out_loops[x]:  # Loop already defined
                                    self.controller.s.append_pipeline_loop(self.loops[x]['type'],
                                                                           self.loops[x]['condition'],
                                                                           "loop" +
                                                                           str(x),
                                                                           parent_loops,
                                                                           list(
                                                                               loop['mod']),
                                                                           xml_parent,
                                                                           self.loops[x]['coord']
                                                                           )
                                    out_loops[x] = 1
                                # parent is the last loop
                                xml_parent = "loop"+str(x)
                                parent_loops.append("loop"+str(x))

                    self.controller.s.append_pipeline_module(self.module_list[i],
                                                             mn[i],
                                                             "",
                                                             {},
                                                             list(
                                                                 mn[values[:, i]]),
                                                             list(
                                                                 mn[values[i, :]]),
                                                             xml_parent,
                                                             [self.canvas_startxy[i], i, self.connections[i]])

            self.controller.s.append_pipeline_module(self.module_list[1],  # Out
                                                     mn[1],
                                                     "",
                                                     {},
                                                     list(mn[values[:, 1]]),
                                                     list(mn[values[1, :]]),
                                                     None,
                                                     [self.canvas_startxy[1], 1, self.connections[1]])
            self.controller.s.write_to_XML()
            self.saved = True
            self.controller._append_to_output(
                "xml_filename", self.save_path.name)

    def upload(self):

        filename = askopenfilename(initialdir=os.getcwd(),
                                   title='Select a file',
                                   defaultextension='.xml',
                                   filetypes=[('XML file', '.xml'),
                                              ('All Files', '*.*')])
        if filename is not None and len(filename) > 0:
            self.reset()

            s = XML_handler()
            s.load_XML(filename)
            # s._print_pretty(s.loaded_modules)
            modules = s.loaded_modules
            modout = modules['Output']
            del modules['Initialiser'], modules['Output'] # They are generated when resetting
            disp_mod = ['Initialiser', 'Output']
            id_mod = [0, 1]

            # Place the modules
            id_mod, disp_mod = self.place_modules(modules, id_mod, disp_mod)
            self.draw_connection(modules, id_mod, disp_mod)
            connect = list(modout['coordinates'][2].keys())
            for p, parent in enumerate(modout['parents']):
                    parent_id = id_mod[np.where(np.array(disp_mod) == parent)[0][0]]
                    out, ins = modout['coordinates'][2][connect[p]].split('-')
                    xout, yout, _ , _ = self.canvas.coords(out[0]+str(parent_id))
                    xins, yins, _, _ =self.canvas.coords(ins[0]+str(1))
                    self.canvas.create_line(
                                xout + self.cr, 
                                yout + self.cr, 
                                xins + self.cr, 
                                yins + self.cr,
                                fill = "red",  width = 2,
                                arrow = tk.LAST, arrowshape = (12,10,5), 
                                tags = ('o'+str(parent_id), 
                                      'o'+str(1), modout['coordinates'][2][connect[p]]))
                    self.out_data.iloc[int(parent_id)][1] = 1
                    self.connections[1][
                        int(parent_id)] = out[0]+str(parent_id) + '-' + ins[0]+str(1)
            self.controller._append_to_output("xml_filename",filename)

    def place_modules(self, modules, id_mod, disp_mod):
        """Places the modules in the dictionary in the canvas.
        :param modules: dict type of modules in the pipeline.
        """
        for key in [key for key, val in modules.items() if type(val) == dict]:
            if modules[key]['class'] == 'loop':
                self.l += 1
                # Extracts numbers from string
                l = int(
                    ''.join(map(str, list(filter(str.isdigit, modules[key]['name'])))))
                x0, y0, x1, y1 = modules[key]['coordinates']
                self.canvas.create_rectangle(x0, y0,
                                             x1, y1,
                                             outline='#4ff07a',
                                             tag='loop-'+str(l))
                text_w = self.controller.pages_font.measure(
                    modules[key]['type']) + 20
                self.canvas.create_text(
                    x0 + text_w/2,
                    y0 + 20,
                    font=self.controller.pages_font,
                    text=modules[key]['type'],
                    tags=('loop-'+str(l), 'type'+'-'+str(l)),
                    justify=tk.CENTER)
                text_w = self.controller.pages_font.measure(
                    modules[key]['condition']) + 20
                self.canvas.create_text(
                    x1 - text_w/2, 
                    y0 + 20, 
                    font = self.controller.pages_font, 
                    text = modules[key]['condition'], 
                    tags = ('loop-'+str(l), 'condition'+'-'+str(l)), 
                    justify = tk.CENTER)
                self.canvas.tag_bind('condition-'+str(l), 
                                 "<Double-1>", self.OnDoubleClick)
                self.canvas.tag_bind('type-'+str(l), 
                                 "<Double-1>", self.OnDoubleClick)
                self.loops.append({'type': modules[key]['type'],
                                   'condition': modules[key]['condition'],
                                   'mod': [],
                                   'coord': (x0, y0,
                                             x1, y1)})
                id_mod, disp_mod = self.place_modules(modules[key], id_mod, disp_mod)
            elif key not in disp_mod:
                # Display module
                self.add_module(key,
                                modules[key]['coordinates'][0][0],
                                modules[key]['coordinates'][0][1], 
                                iid = modules[key]['coordinates'][1])
                self.module_list[-1] = modules[key]['module_type']
                id_mod.append(modules[key]['coordinates'][1])
                disp_mod.append(key)
        self.module_list = [x for _, x in sorted(zip(id_mod, self.module_list))]
        self.canvas_startxy = [x for _, x in sorted(zip(id_mod, self.canvas_startxy))]
        return id_mod, disp_mod

    def draw_connection(self, modules, id_mod, disp_mod):
        for key in [key for key, val in modules.items() if type(val) == dict]:
            if modules[key]['class'] == 'loop':
                self.draw_connection(modules[key], id_mod, disp_mod)
            else:
                # Connect modules
                connect = list(modules[key]['coordinates'][2].keys())
                for p, parent in enumerate(modules[key]['parents']):
                    if not (parent[:4] == 'loop'):
                        # parent_id = modules[parent]['coordinates'][1]
                        parent_id = id_mod[np.where(np.array(disp_mod) == parent)[0][0]]
                        out, ins = modules[key]['coordinates'][2][connect[p]].split('-')
                        xout, yout, _ , _ = self.canvas.coords(out)
                        xins, yins, _, _ = self.canvas.coords(ins)
                        self.canvas.create_line(
                                    xout + self.cr,
                                    yout + self.cr,
                                    xins + self.cr,
                                    yins + self.cr,
                                    fill = "red", width = 2,
                                    arrow = tk.LAST, arrowshape = (12,10,5),
                                    tags = ('o'+str(parent_id),
                                          # 'o'+str(id_mod[-1]), 
                                          modules[key]['coordinates'][2][connect[p]]))
                        self.out_data.iloc[int(parent_id)][int(ins[1:])] = 1
                        self.connections[int(ins[1:])][
                            int(parent_id)] = out[0]+str(out[1:]) + '-' + ins[0]+str(ins[1:])
                    else:
                        self.loops[-1]['mod'].append(key)

    def reset(self):

        if not self.saved:
            msg = messagebox.askyesnocancel(
                'Info', 'Are you sure you want to reset the canvas?')
        else:
            msg = True
        if msg:
            self.canvas.delete(tk.ALL)  # Reset canvas

            if hasattr(self, 'entry'):
                self.entry.destroy()
            if hasattr(self, 'entry1'):
                self.entry1.destroy()
            if hasattr(self, 'entry2'):
                self.entry2.destroy()

            self.canvas_startxy = []
            self.out_data = pd.DataFrame()
            self.connections = {}
            self.modules = 0
            self.module_list = []
            self.module_names = []

            self.add_module('Initialiser', self.width/2, self.h, ini=True)
            self.add_module('Output', self.width/2,
                            self.height - self.h, out=True)

            self.draw = False
            self.loops = []
            self.drawLoop = False
            self.l = 0

    def check_quit(self):

        if not self.saved:
            response = messagebox.askokcancel(
                "Are you sure you want to leave?",
                "Do you want to leave the program without saving?")
            if response:
                if self.save_path not in [None, '']:
                    self.controller.XML.set(True)
                self.controller._show_frame("MainPage")
        else:
            if self.save_path not in [None, '']:
                self.controller.XML.set(True)
            self.controller._show_frame("MainPage")

# class ResizingCanvas(tk.Canvas):
#     def __init__(self,parent,**kwargs):
#         tk.Canvas.__init__(self,parent,**kwargs)
#         self.bind("<Configure>", self.on_resize)
#         self.height = self.winfo_reqheight()
#         self.width = self.winfo_reqwidth()

#     def on_resize(self,event):
#         # determine the ratio of old width/height to new width/height
#         wscale = float(event.width)/self.width
#         hscale = float(event.height)/self.height
#         self.width = event.width
#         self.height = event.height
#         # resize the canvas
#         self.config(width=self.width, height=self.height)
#         # rescale all the objects tagged with the "all" tag
#         self.scale("all",0,0,wscale,hscale)



from vai_lab._import_helper import get_lib_parent_dir

import tkinter as tk
import os
from PIL import Image, ImageTk
from tkinter import messagebox
from tkinter.filedialog import askopenfilename
import pandas as pd

_PLUGIN_READABLE_NAMES = {"start": "default",
                          "start page": "alias"}                    # type:ignore
_PLUGIN_MODULE_OPTIONS = {}                                         # type:ignore
_PLUGIN_REQUIRED_SETTINGS = {}                                      # type:ignore
_PLUGIN_OPTIONAL_SETTINGS = {}                                      # type:ignore


class StartPage(tk.Frame):
    """Main Page for launching GUI methods"""

    def __init__(self, parent, controller):
        " Here we define the main frame displayed upon opening the program."
        " This leads to the different methods to provide feedback."
        super().__init__(parent, bg=parent['bg'])
        self.controller = controller
        self.controller.title('User interaction adaptation')

        # script_dir = os.path.dirname(__file__)
        self.out_data = pd.DataFrame()
        script_dir = get_lib_parent_dir()
        self.my_img1 = ImageTk.PhotoImage(
            Image.open(
                os.path.join(
                    script_dir,
                    "utils",
                    'resources',
                    'Assets',
                    'UFAIcon_name.png')
            ).resize((600, 400))
        )

        self.my_label = tk.Label(self,
                                 image=self.my_img1,
                                 bg=parent['bg'])

        self.my_label.grid(column=0,
                           row=0,
                           rowspan=10,
                           columnspan=4)

        my_label = tk.Label(self,
                            text='Choose feedback method ...',
                            pady=10,
                            font=controller.title_font,
                            bg=parent['bg'],
                            fg='white')

        my_label.grid(column=0,
                      row=11,
                      columnspan=4)

        button_speak = tk.Button(self,
                                 text='Speak',
                                 fg='white',
                                 font=controller.title_font,
                                 bg=parent['bg'],
                                 height=3,
                                 width=20,
                                 command=self.NotImpl
                                 ).grid(column=0, row=59)

        button_write = tk.Button(self,
                                 text='Write',
                                 fg='white',
                                 font=controller.title_font,
                                 bg=parent['bg'],
                                 height=3,
                                 width=20,
                                 command=self.NotImpl
                                 ).grid(column=1, row=59)

        button_manual = tk.Button(self,
                                  text='Manual input',
                                  fg='white',
                                  font=controller.title_font,
                                  bg=parent['bg'],
                                  height=3,
                                  width=20,
                                  command=lambda: controller.show_frame(
                                      "PageManual")
                                  ).grid(column=2, row=59)

        button_canvas = tk.Button(self,
                                  text='Interact with canvas',
                                  fg='white',
                                  font=controller.title_font,
                                  bg=parent['bg'],
                                  height=3,
                                  width=20,
                                  command=lambda: controller.show_frame(
                                      "PageCanvas")
                                  ).grid(column=3, row=59)

        button_upload = tk.Button(self,
                                  text='Upoad file',
                                  fg='white',
                                  font=controller.title_font,
                                  bg=parent['bg'],
                                  height=3,
                                  width=20,
                                  command=self.upload_data
                                  ).grid(column=0, row=60)

    # def light_theme(self):
    #     listbox_tasks.config(bg="white", fg="black")
    #     button_add_task.config(highlightbackground='white')
    #     button_delete_task.config(highlightbackground='white')
    #     button_load_tasks.config(highlightbackground='white')
    #     button_save_tasks.config(highlightbackground='white')
    #     entry_task.config(bg='white', fg='black')

    def NotImpl(self):
        messagebox.showwarning(
            "Error", "This functionality is not implemented yet.")

    def upload_data(self):
        filename = askopenfilename(initialdir=os.getcwd(),
                                   title='Select a file',
                                   defaultextension='.csv',
                                   filetypes=[
            ('Text file', '.txt'),
            ('CSS file', '.css'),
            ('All Files', '*.*')
        ])
        if filename is not None:
            messagebox.showinfo("Success", "Feedback correctly uploaded.")



"""Adds vai_lab root folder to pythonpath assuming this script is 2 levels down.
Hacky patch For testing only, normally you'd have vai_lab in your
OS's PYTHONPATH"""
from os import path

if not __package__:
    import sys
    root_mod = path.dirname(path.dirname(path.dirname(__file__)))
    sys.path.append(root_mod)

from typing import Dict, KeysView, TypeVar

from vai_lab._import_helper import get_lib_parent_dir
from vai_lab.Data.xml_handler import XML_handler
from vai_lab._import_helper import rel_to_abs

import pandas as pd # type: ignore
import numpy as np

DataT = TypeVar("DataT",bound="Data")
class Data:
    
    def __init__(self: DataT) -> None:
        self._lib_base_path = get_lib_parent_dir()
        self._xml_parser = XML_handler()
        self.data: Dict[str, pd.core.frame.DataFrame] = {}

    def _import_csv(self: DataT,
                    filename: str,
                    data_name: str,
                    strip_whitespace: bool = True) -> None:
        """import data directly into DataFrame
        :param filename: str, filename of csv file to be loaded
        :param data_name: str, name of dict key in which data will be stored
        :param strip_whitespace: bool, remove spaces from before & after header names
        TODO: pandas has a lot of inbuilt read functions, including excel - implement
        """
        self.data[data_name] = pd.read_csv(filename,
                                           delimiter=',',
                                           quotechar='|')
        if strip_whitespace:
            self.data[data_name].columns = [c.strip()
                                            for c in self.data[data_name].columns]

    def _import_png(self: DataT,
                    filename: str,
                    data_name: str) -> None:
        """Loads png into PIL.Image class. Adds instance to self.data
        The image is stored as a function (not a matrix - can be added if needed)
        :param filename: str, filename of csv file to be loaded
        :param data_name: str, name of dict key in which data will be stored
        """
        import cv2 # type: ignore
        key = filename.split(path.sep)[-1].split(".")[0]
        self.data[data_name][key] = cv2.imread(filename)

    def _import_dir(self: DataT,
                    folder_dir: str,
                    data_name: str) -> None:
        """Explores folder, and imports all data items recursively
        
        :param folder_dir: str, directory to be explored
        :param data_name: str, name of dict key in which data will be stored
                data_name is "data" overwritten by folder name due to recursion
                This will probably change for user-defined names
        """
        from glob import glob
        if folder_dir[-1] != path.sep:
            folder_dir += path.sep
        # data_name = folder_dir.split(path.sep)[-2]
        files = np.sort(glob(folder_dir + "*"))
        self.data[data_name] = {}
        for f in files:
            self.import_data(f, data_name)

    def _get_ext(self: DataT, path_dir: str) -> str:
        """Extracts extension from path_dir, or check if is dir
        :param path_dir: str, path_dir to be checked
        :returns ext: str, path_dir extension or "dir" if path_dir is directory
        """
        if path.isdir(path_dir):
            return "dir"
        else:
            return path_dir.split(".")[-1]

    def import_data(self: DataT,
                    filename: str,
                    data_name: str = "data") -> None:
        """Import file directly into DataFrame
        Translates relative files to absolute before parsing - not ideal
        Filename to parsing method based on extension name.
        :param filename: str, filename of file to be loaded
        :param data_name: str, name of class variable data will be loaded to
        """
        filename = rel_to_abs(filename).replace(
            "\\", "/").replace("/", path.sep)
        ext = self._get_ext(filename)
        getattr(self, "_import_{0}".format(ext))(filename, data_name)

    def import_data_from_config(self: DataT, config: dict) -> None:
        for c in config.keys():
            self.import_data(config[c], c)

    def append_data_column(self: DataT, col_name: str, data=None) -> None:
        self.data[col_name] = data

    def _parse_data_options(self: DataT) -> None:
        for s in self.loaded_options["data_fields"]:
            self.append_data_column(s)

    def load_data_settings(self: DataT, filename: str) -> None:
        self._xml_parser.load_XML(filename)
        self.loaded_options = self._xml_parser._get_init_data_structure()
        self._parse_data_options()

    def keys(self: DataT) -> KeysView:
        return self.data.keys()

    def __getitem__(self: DataT, key: str) -> pd.core.frame.DataFrame:
        return self.data[key]

    def copy(self: DataT) -> DataT:
        from copy import deepcopy
        return deepcopy(self)

if __name__ == "__main__":
    d = Data()
    # d.load_data_settings("./examples/data_passing_test.xml")
    d.import_data("./examples/crystalDesign")
    print(d["crystalDesign"])
    # print(d["data"]["input"].loc[0:3])
    # dc = d.copy()
    # d["data"]["input"].loc[0:3] = 9
    # print(d["data"]["input"].loc[0:3])
    # print(dc["data"]["input"].loc[0:3])
    # dc.keys()
    # print(dc["input"])
    # d.data_names


# -*- coding: utf-8 -*-

"""Import all submodules inside this folder and expose them with their respective names

Let's look for a more automatic way of doing this when we have more modules
"""
# from .AID_core import AID
# from .UserInterfaceClass import UI
# from . import import_helper



import numpy as np
import pandas as pd
from pathlib import Path  

N = 100
D = 5

X = np.random.rand(N, D)
Y = ((X @ np.random.rand(D, 1) + np.random.normal(0, 2, (N, 1))) > 0).astype(int)

X_tr, Y_tr = pd.DataFrame(X[:int(N*0.6),:]), pd.DataFrame(Y[:int(N*0.6)])
X_tst, Y_tst = pd.DataFrame(X[int(N*0.6):,:]), pd.DataFrame(Y[int(N*0.6):])

filepath = Path('X_tr.csv')
filepath.parent.mkdir(parents=True, exist_ok=True)
X_tr.to_csv(filepath, index = False)
filepath = Path('Y_tr.csv')
filepath.parent.mkdir(parents=True, exist_ok=True)
Y_tr.to_csv(filepath, index = False)
filepath = Path('X_tst.csv')
filepath.parent.mkdir(parents=True, exist_ok=True)
X_tst.to_csv(filepath, index = False)
filepath = Path('Y_tst.csv')
filepath.parent.mkdir(parents=True, exist_ok=True)
Y_tst.to_csv(filepath, index = False)


# -*- coding: utf-8 -*-

"""Import all submodules inside this folder and expose them with their respective names

Let's look for a more automatic way of doing this when we have more modules
"""
from .vai_lab_core import Core
# from .. import *


"""
CLI for running a pipeline from a config file 
or start a GUI if no config file given. 
"""

import argparse
from os.path import abspath

import vai_lab as ai



def parse_args():
    """
    Parse command line arguments
    """

    parser = argparse.ArgumentParser(
                        prog = 'vai_lab',
                        description = 'AI-assisted virtual laboratories',
                        )

    parser.add_argument(
                        '-f', 
                        '--file',
                        type=str,
                        nargs='+',
                        default=None,
                        help='pipeline config file',                        
                        )

    args = parser.parse_args()

    return args

def _config_files_iter(core,files):
    """Iterate over config files and load them into core
    TODO: WIP - need a reset function to avoid instantiating core every time
    """
    if files:
        for f in files:
            core = ai.Core()
            core.load_config_file(abspath(f))
            core.run()


def main():

    # Parse command line arguments
    args = parse_args()

    # Core instance
    core = ai.Core()

    # Load config file if given
    if args.file:
        for i in range(0,len(args.file)):
            args.file[i] = abspath(args.file[i])
        core.load_config_file(args.file)

    # Run pipeline   
    core.run()




if __name__=='__main__':
    
    main()